"""Functionality for creating Deephaven tables."""

import logging
from typing import List, Any, Sequence, Optional, Set
import collections

from deephaven.time import now
import jpy
from deephaven import DynamicTableWriter
from deephaven.table import Table
from deephaven import dtypes
from deephaven.dtypes import DType
import decimal
from decimal import Decimal

from .trace import trace_str


class TableWriter:
    """A writer for logging data to Deephaven dynamic tables.

    Empty strings are logged as None.
    """

    _dtw: DynamicTableWriter
    _string_indices: List[int]
    _receive_time: bool

    def __init__(self, names: List[str], types: List[DType], receive_time: bool = True):
        TableWriter._check_for_duplicate_names(names)
        self.names = names
        self.types = types
        self._receive_time = receive_time

        if receive_time:
            self.names.insert(0, "ReceiveTime")
            self.types.insert(0, dtypes.DateTime)

        col_defs = {name: type for name, type in zip(names, types)}
        self._dtw = DynamicTableWriter(col_defs)
        self._string_indices = [i for (i, t) in enumerate(types) if t == dtypes.string]

    @staticmethod
    def _check_for_duplicate_names(names: List[str]) -> None:
        counts = collections.Counter(names)
        dups = [name for name, count in counts.items() if count > 1]

        if len(dups) > 0:
            raise Exception(f"Duplicate column names: {','.join(dups)}")

    def _check_logged_value_types(self, values: List) -> None:
        for n, t, v in zip(self.names, self.types, values):
            if v is None:
                continue

            if (t is dtypes.string and not isinstance(v, str)) or \
                    (t is dtypes.int64 and not isinstance(v, int)) or \
                    (t is dtypes.float64 and not isinstance(v, float)):
                logging.error(
                    f"TableWriter column type and value type are mismatched: column_name={n} column_type={t} value_type={type(v)} value={v}\n{trace_str()}\n-----")

    def table(self) -> Table:
        """Gets the table data is logged to."""
        return self._dtw.table

    def write_row(self, values: List) -> None:
        """Writes a row of data.  The input values may be modified."""

        if self._receive_time:
            values.insert(0, now())

        for i in range(len(values)):
            if  isinstance(values[i], decimal.Decimal): 
                values[i] = float(values[i])

        self._check_logged_value_types(values)

        for i in self._string_indices:
            if values[i] == "":
                values[i] = None

        try:
            self._dtw.write_row(*values)
        except Exception as e:
            msg = f"Problem logging row:\n"

            for i, v in enumerate(values):
                msg += f"\t{i} {type(v)} {v}\n"

            logging.error(msg)

            raise e


ArrayStringSet = jpy.get_type("io.deephaven.stringset.ArrayStringSet")

_unmapped_values_already_logged:Set[str] = set()

def map_values(value, map, default=lambda v: f"UNKNOWN({v})") -> Any:
    """ Maps one set of values to another.  A default value is used if the value is not in the map. """

    if value is None:
        return None

    try:
        return map[value]
    except KeyError:
        msg = f"Unmapped value.  Please file an issue at https://github.com/deephaven-examples/deephaven-ib/issues: '{value}'\n{trace_str()}\n-----"

        if msg not in _unmapped_values_already_logged:
            _unmapped_values_already_logged.add(msg)
            logging.error(msg)

        return default(value)


def to_string_val(value) -> Optional[str]:
    """ Converts a value to a string. """

    if value is None:
        return None

    return str(value)


def to_string_set(value: Sequence) -> Optional[ArrayStringSet]:
    """ Converts an iterable to a string set. """

    if value is None:
        return None

    return ArrayStringSet(list({to_string_val(v) for v in value}))



"""Functionality for working with time in Deephaven and IB."""

import jpy
import deephaven.time as dtime
from deephaven.dtypes import DateTime

_SimpleDateFormat = jpy.get_type("java.text.SimpleDateFormat")
_TimeZone = jpy.get_type("java.util.TimeZone")

_ib_date_time_pattern_sec = "yyyyMMdd HH:mm:ss"
_ib_date_time_pattern_subsec = "yyyyMMdd HH:mm:ss.S"
_ib_date_time_patterns = [
    "yyyyMMdd HH:mm:ss.S",
    "yyyy-MM-dd HH:mm:ss.S",
    "yyyyMMdd HH:mm:ss",
    "yyyy-MM-dd HH:mm:ss",
]
_ib_date_time_formatter_sec = _SimpleDateFormat(_ib_date_time_pattern_sec)
_ib_date_time_formatter_sec.setTimeZone(_TimeZone.getTimeZone("US/Eastern"))
_ib_date_time_formatter_subsec = _SimpleDateFormat(_ib_date_time_pattern_subsec)
_ib_date_time_formatter_subsec.setTimeZone(_TimeZone.getTimeZone("US/Eastern"))
_ib_date_time_formatters = [_SimpleDateFormat(pattern) for pattern in _ib_date_time_patterns]

for _f in _ib_date_time_formatters:
    _f.setTimeZone(_TimeZone.getTimeZone("US/Eastern"))

def dh_to_ib_datetime(time: DateTime, sub_sec: bool = True) -> str:
    """Convert a DH DateTime to an IB timestamp string.

    Args:
        time (DateTime): time
        sub_sec (bool): true to return subsecond resolution and false otherwise.
    """

    if time is None:
        return ""

    if sub_sec:
        return _ib_date_time_formatter_subsec.format(time.getDate()) + " US/Eastern"
    else:
        return _ib_date_time_formatter_sec.format(time.getDate()) + " US/Eastern"


def ib_to_dh_datetime(time: str) -> DateTime:
    """Convert an IB timestamp to a DH DateTime."""

    if time is None:
        return None

    exceptions = []

    for formatter in _ib_date_time_formatters:
        try:
            return DateTime.j_type.of(formatter.parse(time).toInstant())
        except Exception as e:
            exceptions.append(e)
            pass

    raise Exception(f"Unable to parse time '{time}'", exceptions)


def unix_sec_to_dh_datetime(time: int) -> DateTime:
    """Convert Unix seconds since the epoch to a DH DateTime."""

    if time is None:
        return None

    return DateTime(int(time) * dtime.SECOND)



"""Functionality for working with error codes."""

from typing import Dict, Tuple

import pandas as pd


def load_error_codes() -> Tuple[Dict[int, str], Dict[int, str]]:
    """Load dictionaries of error code messages and notes."""

    html_tables = pd.read_html('https://interactivebrokers.github.io/tws-api/message_codes.html')

    error_messages = {}
    error_notes = {}

    for df in html_tables:
        try:
            df = df.fillna('')
            codes = df['Code']
            messages = df['TWS message']
            notes = df['Additional notes']

            for code, message in zip(codes, messages):
                error_messages[code] = message

            for code, note in zip(codes, notes):
                error_notes[code] = note
        except KeyError:
            pass

    overrides = {
        0: "Warning: Approaching max rate of 50 messages per second",
        504: "Not connected",
        502: "Couldn't connect to TWS. Confirm that 'Enable ActiveX and Socket EClients' is enabled and connection port is the same as 'Socket Port' on the TWS 'Edit->Global Configuration...->API->Settings' menu. Live Trading ports: TWS: 7496; IB Gateway: 4001. Simulated Trading ports for new installations of version 954.1 or newer:  TWS: 7497; IB Gateway: 4002",
        2113: "The order size for Bonds (Bills) is entered as a nominal par value of the order, and must be a multiple",
        2157: "Sec-def data farm connection is broken:secdefil",
        10089: "Requested market data requires additional subscription for API.See link in 'Market Data Connections' dialog for more details.",
        10147: "OrderId that needs to be cancelled is not found.",
        10168: "Requested market data is not subscribed.Delayed market data is not enabled.",
        10172: "Failed to request news article: No data available",
        10187: "Failed to request historical ticks",
        10189: "Failed to request tick-by-tick data",
        10190: "Maxnumber of tick-by-tick requests has been reached.",
    }

    for k, v in overrides.items():
        if k not in error_messages:
            error_messages[k] = v
            error_notes[k] = ""

    return error_messages, error_notes



import os

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

version = os.getenv("DH_IB_VERSION")

if not version:
    raise Exception("Version must be set via the DH_IB_VERSION environment varialble.")

setuptools.setup(
    name="deephaven_ib",
    version=version,
    author="David R. (Chip) Kent IV",
    author_email="chipkent@deephaven.io",
    description="An Interactive Brokers integration for Deephaven",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/deephaven-examples/deephaven-ib",
    project_urls={
        "Deephaven": "https://deephaven.io",
        "Deephaven GitHub": "https://github.com/deephaven/deephaven-core",
        "GitHub Issues": "https://github.com/deephaven-examples/deephaven-ib/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Financial and Insurance Industry",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: Office/Business :: Financial :: Investment",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=[
        "deephaven-server==0.17.0",
        "pandas",
        "ibapi==10.16.1",
        "lxml",
        "ratelimit",
    ],
)

